SystemOrganization addCategory: #Unification!

!AlgebraicDataType methodsFor: '*Unification' stamp: 'fbs 5/30/2011 20:20'!
findSolutionWithUnifier: aDictionary in: termRelation
	| newUnifier |
	newUnifier := aDictionary copy.
	self children do:
		[:each | | d |
		d := (termRelation classFor: each)
			findSolutionWithUnifier: newUnifier in: termRelation.
		newUnifier addAll: d].
	^ newUnifier.! !

!AlgebraicDataType methodsFor: '*Unification' stamp: 'fbs 5/25/2011 22:31'!
isMetaVariable
	^ self unapply first = #Variable.! !

!AlgebraicDataType methodsFor: '*Unification' stamp: 'fbs 6/2/2011 13:51'!
unificationClosureWith: anObject in: termRelation
	anObject unificationClosureWithAlgebraic: self in: termRelation.! !

!AlgebraicDataType methodsFor: '*Unification' stamp: 'fbs 6/2/2011 13:54'!
unificationClosureWithAlgebraic: anObject in: termRelation
	"s = f(s1, ..., sm), t = g(t1, ..., tn) where m, n > 0"
	(self type = anObject type) ifFalse: [UnificationFailure signal. "f ~= g"].
	(self arity = anObject arity) ifFalse: [UnificationFailure signal. "m ~= n"].
	self children with: anObject children do:
		[:si :ti | (termRelation classFor: si) unificationClosureWith: (termRelation classFor: ti) in: termRelation].! !

AlgebraicDataType subclass: #MetaVariable
	instanceVariableNames: 'name'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Unification'!

!MetaVariable commentStamp: 'fbs 5/24/2011 11:04' prior: 0!
A MetaVariable is a reified variable, that is, an object representing a variable. It is a placeholder in a structure that says "we don't know what the value is of this part of the structure, but we can give it a name."!

!MetaVariable class methodsFor: 'instance creation' stamp: 'fbs 5/24/2011 07:31'!
named: aSymbol
	^ self new name: aSymbol.! !

!MetaVariable methodsFor: 'accessing' stamp: 'fbs 5/24/2011 22:06'!
children
	^ #().! !

!MetaVariable methodsFor: 'testing' stamp: 'fbs 5/24/2011 22:05'!
hasChildren
	"Even though unapplied a variable looks like it has a parameter, that's just its name."
	^ false.! !

!MetaVariable methodsFor: 'accessing' stamp: 'fbs 5/24/2011 07:30'!
name
	^ name.! !

!MetaVariable methodsFor: 'private' stamp: 'fbs 5/24/2011 07:32'!
name: aSymbol
	name := aSymbol.! !

!MetaVariable methodsFor: 'destructuring' stamp: 'fbs 5/24/2011 07:34'!
unapply
	^ {#Variable. name.}! !

Error subclass: #UnificationFailure
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Unification'!

!UnificationFailure commentStamp: 'fbs 5/24/2011 17:35' prior: 0!
I represent the inability to unify two patterns.!

!Object methodsFor: '*Unification' stamp: 'fbs 5/25/2011 22:33'!
isMetaVariable
	^ false.! !

!Object methodsFor: '*Unification' stamp: 'fbs 7/23/2011 11:25'!
matchOneOf: aBlockAssociationCollection
	"The elements of aBlockAssociationCollection are associations between blocks.
	 Answer the evaluated value of the first association in aBlockAssociationCollection
	 whose evaluated key unifies with the receiver.  If no match is found, report an error."

	^ self matchOneOf: aBlockAssociationCollection otherwise: [self caseError].! !

!Object methodsFor: '*Unification' stamp: 'fbs 7/22/2011 08:43'!
matchOneOf: aBlockAssociationCollection otherwise: aBlock
	"The elements of aBlockAssociationCollection are associations between blocks.
	 Answer the evaluated value of the first association in aBlockAssociationCollection
	 whose evaluated key unifies with the receiver.  If no match is found, answer the result
	 of evaluating aBlock."
	| mgu |
	aBlockAssociationCollection associationsDo:
		[:assoc |
			[mgu := assoc key value =? self.
			mgu isEmpty
				ifFalse: [^ assoc value cull: mgu]] on: UnificationFailure do: ["nothing"]].
	^ aBlock value! !

!Object methodsFor: '*Unification' stamp: 'fbs 7/26/2011 21:44'!
unificationClosureWith: anObject in: termRelation
	(self = anObject)
		ifTrue: [termRelation union: self with: anObject]
		ifFalse: ["s ~= t" UnificationFailure signal]! !

!Object methodsFor: '*Unification' stamp: 'fbs 6/2/2011 13:50'!
unificationClosureWithAlgebraic: anObject in: termRelation
	UnificationFailure signal.! !

!Object methodsFor: '*Unification' stamp: 'fbs 5/26/2011 20:42'!
unifyWith: anObject
	^ Unifier unify: self with: anObject.! !

Object subclass: #TermRelation
	instanceVariableNames: 'map'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Unification'!

!TermRelation commentStamp: 'fbs 5/28/2011 15:43' prior: 0!
I represent the term relation we build up as part of unifying two structures.!

!TermRelation methodsFor: 'accessing' stamp: 'fbs 5/28/2011 18:50'!
classFor: anObject
	^ map
		at: anObject
		ifAbsentPut: [UnificationEquivalenceClass decorate: anObject].! !

!TermRelation methodsFor: 'initialize-release' stamp: 'fbs 5/28/2011 15:52'!
initialize
	super initialize.
	map := Dictionary new.! !

!TermRelation methodsFor: 'as yet unclassified' stamp: 'fbs 7/26/2011 21:44'!
union: thisObject with: thatObject
	(self classFor: thisObject) findRepresentative schemaTerm
			union: (self classFor: thatObject) findRepresentative schemaTerm! !

Object subclass: #UnificationEquivalenceClass
	instanceVariableNames: 'classPointer classSize schemaTerm visited acyclic vars node'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Unification'!

!UnificationEquivalenceClass commentStamp: 'fbs 5/28/2011 15:45' prior: 0!
I decorate the nodes undergoing unification. Unifier associates a node with a UnificationDecoration to allow the unification state to remain separate from the immutable nodes.

* classPointer points to my equivalence class; if classPointer = self then I am the representative for my class.
* classSize is an optimisation: when we merge two equivalence classes, we merge the smaller class with the larger class.
* node points to the thing that I decorate.!

!UnificationEquivalenceClass class methodsFor: 'instance creation' stamp: 'fbs 5/25/2011 21:16'!
decorate: anObject
	^ self new decorate: anObject.! !

!UnificationEquivalenceClass methodsFor: 'accessing' stamp: 'fbs 5/28/2011 15:46'!
classPointer: anEquivClass
	classPointer := anEquivClass.! !

!UnificationEquivalenceClass methodsFor: 'accessing' stamp: 'fbs 5/25/2011 17:58'!
classSize
	^ classSize.! !

!UnificationEquivalenceClass methodsFor: 'initialize-release' stamp: 'fbs 5/25/2011 22:31'!
decorate: aNode
	self node: aNode.
	vars := aNode isMetaVariable
		ifTrue: [aNode asOrderedCollection]
		ifFalse: [OrderedCollection new].
	
	acyclic := false.
	classPointer := self. "All nodes start as the sole element in an equivalence class; this makes each node a representative of its class."
	schemaTerm := self.
	classSize := 1.
	visited := false.! !

!UnificationEquivalenceClass methodsFor: 'unification' stamp: 'fbs 5/29/2011 13:44'!
findRepresentative
	"Walk up the tree representing this equivalence class to find the representative. As a side effect, compress the path to the representative."
	| result |
	self isRepresentative ifTrue: [^ self].
	
	result := classPointer findRepresentative.
	classPointer := result.
	^ result.! !

!UnificationEquivalenceClass methodsFor: 'unification' stamp: 'fbs 5/29/2011 13:39'!
findSolutionWithUnifier: aDictionary in: termRelation
	^ (self isSchemaTerm)
		ifTrue: [self privFindSolutionWithUnifier: aDictionary in: termRelation]
		ifFalse: [self findRepresentative schemaTerm findSolutionWithUnifier: aDictionary in: termRelation]! !

!UnificationEquivalenceClass methodsFor: 'testing' stamp: 'fbs 5/29/2011 13:44'!
isRepresentative
	^ classPointer = self! !

!UnificationEquivalenceClass methodsFor: 'testing' stamp: 'fbs 5/29/2011 13:39'!
isSchemaTerm
	^ self = self findRepresentative schemaTerm.! !

!UnificationEquivalenceClass methodsFor: 'accessing' stamp: 'fbs 5/25/2011 20:07'!
node
	^ node.! !

!UnificationEquivalenceClass methodsFor: 'accessing' stamp: 'fbs 5/25/2011 20:07'!
node: anObject
	node := anObject.! !

!UnificationEquivalenceClass methodsFor: 'unification' stamp: 'fbs 5/30/2011 20:21'!
privFindSolutionWithUnifier: aDictionary in: termRelation
	"Assume that we are the schema term for our equivalence class."
	| newUnifier |
	acyclic ifTrue: [^aDictionary "s is not part of a cycle."].
	
	visited ifTrue: [UnificationFailure signal. "Cycle detected."].

	newUnifier := aDictionary copy.	
	visited := true.
	(node isAlgebraic) ifTrue:
		[newUnifier addAll: (node findSolutionWithUnifier: newUnifier in: termRelation)].
	visited := false.
	
	acyclic := true.
	self findRepresentative vars do:
		[:x | (x ~= node) ifTrue: [newUnifier at: x put: node]].
	
	^ newUnifier.! !

!UnificationEquivalenceClass methodsFor: 'accessing' stamp: 'fbs 5/25/2011 17:58'!
schemaTerm
	^ schemaTerm.! !

!UnificationEquivalenceClass methodsFor: 'unification' stamp: 'fbs 6/2/2011 13:54'!
unificationClosureWith: rightEquivClass in: termRelation
	| left right sTerm t tTerm |
	t := rightEquivClass findRepresentative.
	self isRepresentative ifFalse: [^ self findRepresentative unificationClosureWith: t in: termRelation].

	(self = t) ifTrue: [^ self. "Nothing to do"].
		
	sTerm := schemaTerm.
	tTerm := t schemaTerm.
	left := sTerm node.
	right := tTerm node.
	
	(left isMetaVariable or: [right isMetaVariable])
		ifTrue: [^ self union: t].

	"s, t are both terms."
	left unificationClosureWith: right in: termRelation.
	sTerm union: tTerm.! !

!UnificationEquivalenceClass methodsFor: 'unification' stamp: 'fbs 5/30/2011 09:48'!
union: anEquivClass
	"Given myself, a representative, and anEquivClass, another representative, merge the two equivalence classes by folding the smaller into the larger."
	(self classSize >= anEquivClass classSize)
		ifFalse: [^ anEquivClass union: self].
	
	classSize := classSize + anEquivClass classSize.
	vars addAll: anEquivClass vars.
	schemaTerm node isMetaVariable ifTrue: [schemaTerm := anEquivClass schemaTerm].
	anEquivClass classPointer: self.! !

!UnificationEquivalenceClass methodsFor: 'accessing' stamp: 'fbs 5/25/2011 19:47'!
vars
	^ vars! !

Object subclass: #Unifier
	instanceVariableNames: 'termRelation'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Unification'!

!Unifier commentStamp: 'fbs 7/23/2011 11:32' prior: 0!
I attempt to unify two structures, that is, find mappings between the MetaVariables (if any) within the two structures such that the structures represent the same structure.

I return a Dictionary containing the most general unifier (which might be empty if the two structures are identical), or I raise a UnificationFailure.!

!Unifier class methodsFor: 'instance creation' stamp: 'fbs 5/25/2011 07:12'!
unify: left with: right
	^ self new unify: left with: right.! !

!Unifier methodsFor: 'evaluating' stamp: 'fbs 5/28/2011 19:53'!
unify: s with: t
	termRelation := TermRelation new.
	(termRelation classFor: s) unificationClosureWith: (termRelation classFor: t) in: termRelation.
	^ (termRelation classFor: s) findSolutionWithUnifier: Dictionary new in: termRelation.! !

!Symbol methodsFor: '*Unification' stamp: 'fbs 7/22/2011 13:18'!
asVariable
	^ MetaVariable named: self.! !
